# Spec version 0.4.0
# This file is auto-generated by setup.py

from .base_types import AnnotationTypesBase

class AnnotationTypes(AnnotationTypesBase):
    """
    This class contains the CLAMS annotation types 
    defined in the spec version 0.4.0 as class variables. 
    """
    Annotation = AnnotationTypesBase('http://mmif.clams.ai/0.4.0/vocabulary/Annotation')
    Region = AnnotationTypesBase('http://mmif.clams.ai/0.4.0/vocabulary/Region')
    TimePoint = AnnotationTypesBase('http://mmif.clams.ai/0.4.0/vocabulary/TimePoint')
    Interval = AnnotationTypesBase('http://mmif.clams.ai/0.4.0/vocabulary/Interval')
    Span = AnnotationTypesBase('http://mmif.clams.ai/0.4.0/vocabulary/Span')
    TimeFrame = AnnotationTypesBase('http://mmif.clams.ai/0.4.0/vocabulary/TimeFrame')
    Chapter = AnnotationTypesBase('http://mmif.clams.ai/0.4.0/vocabulary/Chapter')
    Polygon = AnnotationTypesBase('http://mmif.clams.ai/0.4.0/vocabulary/Polygon')
    BoundingBox = AnnotationTypesBase('http://mmif.clams.ai/0.4.0/vocabulary/BoundingBox')
    VideoObject = AnnotationTypesBase('http://mmif.clams.ai/0.4.0/vocabulary/VideoObject')
    Relation = AnnotationTypesBase('http://mmif.clams.ai/0.4.0/vocabulary/Relation')
    Alignment = AnnotationTypesBase('http://mmif.clams.ai/0.4.0/vocabulary/Alignment')
