# Spec version 0.4.0
# This file is auto-generated by setup.py

from .base_types import DocumentTypesBase


class DocumentTypes(DocumentTypesBase):
    """
    This class contains the CLAMS document types 
    defined in the spec version 0.4.0 as class variables. 
    """
    Document = DocumentTypesBase('http://mmif.clams.ai/0.4.0/vocabulary/Document')
    VideoDocument = DocumentTypesBase('http://mmif.clams.ai/0.4.0/vocabulary/VideoDocument')
    AudioDocument = DocumentTypesBase('http://mmif.clams.ai/0.4.0/vocabulary/AudioDocument')
    ImageDocument = DocumentTypesBase('http://mmif.clams.ai/0.4.0/vocabulary/ImageDocument')
    TextDocument = DocumentTypesBase('http://mmif.clams.ai/0.4.0/vocabulary/TextDocument')
